from unittest import mock

import pytest

from jsm_user_services.support.local_threading_utils import add_to_local_threading
from jsm_user_services.support.local_threading_utils import remove_from_local_threading
from jsm_user_services.support.logging_utils import UserSessionIDFilter


@pytest.fixture
def session_id_filter() -> UserSessionIDFilter:
    return UserSessionIDFilter()


def test_should_set_record_session_id_to_empty_when_user_session_id_isnt_set(session_id_filter: UserSessionIDFilter):
    remove_from_local_threading("user_session_id")

    mocked_record = mock.MagicMock()

    result = session_id_filter.filter(mocked_record)

    assert result
    assert mocked_record.session_id == UserSessionIDFilter.EMPTY_FIELD


def test_should_set_record_session_id_to_user_session_id(session_id_filter: UserSessionIDFilter):
    add_to_local_threading("user_session_id", "abc")

    mocked_record = mock.MagicMock()

    result = session_id_filter.filter(mocked_record)

    assert result
    assert mocked_record.session_id == "abc"



from jsm_user_services.support.http_utils import get_response_body
import json
from typing import Dict

import httpretty
import pytest
import requests
from requests import Response

from jsm_user_services.services.email_validations import check_response_payload
from jsm_user_services.services.email_validations import check_response_payload_error
from jsm_user_services.support.email_utils import perform_email_validators
from jsm_user_services.support.email_utils import validate_format_email
from jsm_user_services.support.settings_utils import get_from_settings_or_raise_missing_config

EVEREST_API_URL = get_from_settings_or_raise_missing_config(
    "EVEREST_API_URL", "https://api.everest.validity.com/api/2.0/validation/addresses"
)

class TestValidateEmail:

    invalid_emails = [
        "ayron.araujo",
        "loiraojuntossomosmais.io",
        "leydsonjuntossomosmais.lol",
        "aldojuntossomosmais.com.br",
        "ayron41.@gmail.com",
        "emailqualquer@@juntossomosmais.com.br",
        "???@outlook.com",
        "",
    ]

    valid_emails = [
        "ayron41@gmail.com",
        "ayron.araujo@juntossomosmais.com.br",
        "ayron.style@hotmail.com",
    ]

    @pytest.mark.parametrize("email", invalid_emails)
    def test_should_return_false_when_email_is_invalid(self, email: str):

        assert perform_email_validators(
            [
                validate_format_email
            ],
            email
        ) is False

    @pytest.mark.parametrize("email", valid_emails)
    def test_should_return_true_when_email_is_valid(self, email: str):

        assert perform_email_validators(
            [
                validate_format_email
            ],
            email
        ) is True


class TestValidateEmailRequest:

    @httpretty.activate
    def test_should_retry_request_when_status_code_returned_is_504(self):

        email: str = "ayron.araujo@juntossomosmais.com.br"
        url: str = f"{EVEREST_API_URL}/{email}"
        mocked_error_response = {"status": "Gateway Timeout"}
        mocked_response = {"results": {"status": "risky"}}
        error_timeout_httpretty = httpretty.Response(body=json.dumps(mocked_error_response), status=504)
        success_response_httpretty = httpretty.Response(body=json.dumps(mocked_response), status=200)
        httpretty.register_uri(
            httpretty.GET,
            url,
            responses=[
                error_timeout_httpretty,
                error_timeout_httpretty,
                error_timeout_httpretty,
                success_response_httpretty,
            ]
        )
        response1: Response = requests.get(url)
        payload1 = get_response_body(response1)
        assert check_response_payload(payload1) is False

        response2: Response = requests.get(url)
        payload2 = get_response_body(response2)
        assert check_response_payload(payload2) is False

        response3: Response = requests.get(url)
        payload3 = get_response_body(response3)
        assert check_response_payload(payload3) is False

        response4: Response = requests.get(url)
        payload4 = get_response_body(response4)
        assert check_response_payload(payload4) is True

    @httpretty.activate
    def test_should_return_message_when_email_is_valid_and_result_is_none(self):

        email: str = "ayron.araujo@juntossomosmais.com.br"
        url: str = f"{EVEREST_API_URL}/{email}"
        mocked_response = None
        httpretty.register_uri(
            httpretty.GET,
            url,
            responses=[
                httpretty.Response(status=504, body=json.dumps(mocked_response))
            ],
        )
        response: Response = requests.get(url)
        payload = get_response_body(response)
        assert check_response_payload_error(payload) == "Everest is out!"

    @httpretty.activate
    def test_should_return_message_when_email_is_valid_and_result_is_empty(self):

        email: str = "ayron.araujo@juntossomosmais.com.br"
        url: str = f"{EVEREST_API_URL}/{email}"
        mocked_response = ""
        httpretty.register_uri(
            httpretty.GET,
            url,
            responses=[
                httpretty.Response(status=504, body=json.dumps(mocked_response))
            ],
        )
        response: Response = requests.get(url)
        payload = get_response_body(response)
        assert check_response_payload_error(payload) == "Everest is out!"

    @httpretty.activate
    def test_should_return_message_when_response_dict_is_empty(self):

        email: str = "ayron.araujo@juntossomosmais.com.br"
        url: str = f"{EVEREST_API_URL}/{email}"
        mocked_response = {}
        httpretty.register_uri(
            httpretty.GET,
            url,
            responses=[
                httpretty.Response(status=504, body=json.dumps(mocked_response))
            ],
        )
        response: Response = requests.get(url)
        payload = get_response_body(response)
        assert check_response_payload_error(payload) == "Everest is out!"

    @httpretty.activate
    def test_should_return_true_when_status_is_valid(self):

        mocked_response = {"results": {"status": "valid"}}
        email: str = "ayron.araujo@juntossomosmais.com.br"
        url: str = f"{EVEREST_API_URL}/{email}"
        httpretty.register_uri(
            httpretty.GET,
            url,
            responses=[
                httpretty.Response(body=json.dumps(mocked_response), status=200)
            ],
        )
        response: Response = requests.get(url)
        payload = get_response_body(response)
        assert check_response_payload(payload) is True

    @httpretty.activate
    def test_should_return_false_when_status_is_invalid(self):

        mocked_response = {"results": {"status": "invalid"}}
        email: str = "ayron.araujo@juntossomosmais.com.br"
        url: str = f"{EVEREST_API_URL}/{email}"
        httpretty.register_uri(
            httpretty.GET,
            url,
            responses=[
                httpretty.Response(body=json.dumps(mocked_response), status=200)
            ],
        )
        response: Response = requests.get(url)
        payload = get_response_body(response)
        assert check_response_payload(payload) is False

    @httpretty.activate
    def test_should_return_false_when_response_dict_is_empty(self):

        mocked_response = {}
        email: str = "ayron.araujo@juntossomosmais.com.br"
        url: str = f"{EVEREST_API_URL}/{email}"
        httpretty.register_uri(
            httpretty.GET,
            url,
            responses=[
                httpretty.Response(body=json.dumps(mocked_response), status=200)
            ],
        )
        response: Response = requests.get(url)
        payload = get_response_body(response)
        assert check_response_payload(payload) is False



"""
Module with tests for the DRF tools.
"""
from unittest import mock

import pytest
from django.test import TestCase
from requests.exceptions import HTTPError
from requests_mock import Mocker
from rest_framework.exceptions import APIException
from rest_framework.viewsets import ViewSet

from jsm_user_services.drf_tools.permissions import GoogleRecaptchaPermission
from jsm_user_services.drf_tools.permissions import IsUserAllowedToPermission
from jsm_user_services.drf_tools.permissions import RoleBasedPermission
from jsm_user_services.drf_tools.permissions import StatusBasedPermission
from jsm_user_services.exception import IncorrectTypePermissionConfiguration


class TestDRFPermissionClasses(TestCase):
    def setUp(self):
        self.role_based_permission = RoleBasedPermission()
        self.status_based_permission = StatusBasedPermission()
        self.google_permission_recaptcha = GoogleRecaptchaPermission()
        self.allowed_to_based_permission = IsUserAllowedToPermission

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_role_based_permission_should_validate_request_data_and_append_it_to_request(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        del request.user_data
        append_user_data_to_request = True
        remote_user_data = {"name": "Igor G Peternella", "roles": ["professional"]}

        # get_user_data_from_server answers with an http 200 status
        mocked_get_user_data_from_server.return_value = remote_user_data
        user_data = self.role_based_permission._validate_request_against_user_service(
            request, append_user_data_to_request
        )

        expected_user_data = remote_user_data

        self.assertDictEqual(user_data, expected_user_data)
        self.assertDictEqual(request.user_data, expected_user_data)

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_role_based_permission_should_return_empty_dict_upon_401_from_user_service(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        del request.user_data
        append_user_data_to_request = True

        # get_user_data_from_server answers with an http 401 status
        error = HTTPError("Unauthorized error!")
        response = mock.MagicMock()
        response.status_code = 401
        setattr(error, "response", response)

        mocked_get_user_data_from_server.side_effect = error

        user_data = self.role_based_permission._validate_request_against_user_service(
            request, append_user_data_to_request
        )

        expected_user_data = {}

        self.assertDictEqual(user_data, expected_user_data)

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_role_based_permission_should_reraise_exception_when_user_service_returns_500(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        del request.user_data
        append_user_data_to_request = True

        # get_user_data_from_server answers with an http 500 status
        error = HTTPError("A 500 was returned!")
        response = mock.MagicMock()
        response.status_code = 500
        setattr(error, "response", response)

        mocked_get_user_data_from_server.side_effect = error

        with pytest.raises(HTTPError) as e:
            self.role_based_permission._validate_request_against_user_service(request, append_user_data_to_request)
            assert str(e.value) == "A 500 was returned!"

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_permission_shouldnt_call_get_user_data_from_server_when_user_data_is_already_appended_to_request(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        remote_user_data = {"name": "Remote User", "roles": ["professional"]}
        request.user_data = remote_user_data

        user_data = self.role_based_permission._validate_request_against_user_service(request)

        self.assertDictEqual(user_data, remote_user_data)
        mocked_get_user_data_from_server.assert_not_called()

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_false_if_user_data_is_an_empty_dict(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {}

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["professional"])
        self.assertFalse(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_false_if_user_data_has_disallowed_role(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "roles": ["owner"]  # only 'professional' role is allowed
        }

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["professional"])
        self.assertFalse(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_true_if_user_data_has_allowed_role(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "roles": ["professional"]  # only 'professional' role is allowed
        }

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["professional"])
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_true_if_user_data_has_allowed_role_when_more_than_one_role_is_allowed(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "roles": ["manager"]  # only 'manager' and 'employee' roles are allowed
        }

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["employee", "manager"])
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_true_if_user_data_has_allowed_status(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "status": "active"
        }  # only 'active' status is allowed
        permission_rslt = self.status_based_permission._validate_user_status(request, allowed_status=["active"])
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_true_if_user_data_has_allowed_status_when_more_than_one_status_is_allowed(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "status": "pending-validation"
        }  # only 'active' and 'pending-validation' status are allowed

        permission_rslt = self.status_based_permission._validate_user_status(
            request, allowed_status=["active", "pending-validation"]
        )
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_false_if_user_data_has_disallowed_status(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "status": "blocked"
        }  # only pending-validation' status is allowed

        permission_rslt = self.status_based_permission._validate_user_status(
            request, allowed_status=["pending-validation"]
        )
        self.assertFalse(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_false_if_user_data_is_an_empty_dict(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {}

        permission_rslt = self.status_based_permission._validate_user_status(request, allowed_status=["active"])
        self.assertFalse(permission_rslt)

    @mock.patch(
        "jsm_user_services.drf_tools.permissions.IsUserAllowedToPermission._validate_request_against_user_service"
    )
    def test_should_return_false_if_user_data_has_redeem_as_not_allowed_to_parameter_and_its_not_permitted(
        self, mocked_validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem"]}

        permission_rslt = self.allowed_to_based_permission(["redeem"]).has_permission(request, None)
        self.assertFalse(permission_rslt)

    @mock.patch(
        "jsm_user_services.drf_tools.permissions.IsUserAllowedToPermission._validate_request_against_user_service"
    )
    def test_should_return_false_if_user_data_full_match_with_not_allowed_to_parameter(
        self, mocked_validate_request_against_user_service
    ):
        request = mock.MagicMock()
        # user not_allowed_to with one parameter
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem"]}
        permission_rslt = self.allowed_to_based_permission(["redeem"], full_match=True).has_permission(request, None)
        self.assertFalse(permission_rslt)

        # user not_allowed_to with two parameters
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem", "indicate"]}
        permission_rslt = self.allowed_to_based_permission(["redeem", "indicate"], full_match=True).has_permission(
            request, None
        )
        self.assertFalse(permission_rslt)

        # user not_allowed_to with three parameters
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["score", "indicate", "redeem"]}
        permission_rslt = self.allowed_to_based_permission(["redeem", "indicate"], full_match=True).has_permission(
            request, None
        )
        self.assertFalse(permission_rslt)

        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem", "indicate", "score"]}
        permission_rslt = self.allowed_to_based_permission(
            ["redeem", "indicate", "score"], full_match=True
        ).has_permission(request, None)
        self.assertFalse(permission_rslt)

    @mock.patch(
        "jsm_user_services.drf_tools.permissions.IsUserAllowedToPermission._validate_request_against_user_service"
    )
    def test_should_return_true_if_user_data_do_not_full_match_with_not_allowed_to_parameter(
        self, mocked_validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem"]}
        permission_rslt = self.allowed_to_based_permission(["redeem", "indicate"], full_match=True).has_permission(
            request, None
        )
        self.assertTrue(permission_rslt)

        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem", "score"]}
        permission_rslt = self.allowed_to_based_permission(["redeem", "indicate"], full_match=True).has_permission(
            request, None
        )
        self.assertTrue(permission_rslt)

        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem", "indicate"]}
        permission_rslt = self.allowed_to_based_permission(
            ["redeem", "indicate", "score"], full_match=True
        ).has_permission(request, None)
        self.assertTrue(permission_rslt)

    @mock.patch(
        "jsm_user_services.drf_tools.permissions.IsUserAllowedToPermission._validate_request_against_user_service"
    )
    def test_should_raise_exception_if_class_parameter_not_allowed_to_do_not_be_a_list(
        self, mocked_validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem"]}

        with self.assertRaises(IncorrectTypePermissionConfiguration) as context:
            self.allowed_to_based_permission("redeem").has_permission(request, None)
        self.assertTrue("Expected type 'List[str]', got" in context.exception.args[0])

    @mock.patch(
        "jsm_user_services.drf_tools.permissions.IsUserAllowedToPermission._validate_request_against_user_service"
    )
    def test_should_return_false_if_user_data_has_at_list_one_not_allowed_to_parameter(
        self, mocked_validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem"]}

        # full match default equals to False

        permission_rslt = self.allowed_to_based_permission(["redeem", "indication"]).has_permission(request, None)
        self.assertFalse(permission_rslt)

        permission_rslt = self.allowed_to_based_permission(["indication", "redeem"]).has_permission(request, None)
        self.assertFalse(permission_rslt)

        permission_rslt = self.allowed_to_based_permission(["indication", "redeem", "score"]).has_permission(
            request, None
        )
        self.assertFalse(permission_rslt)

        permission_rslt = self.allowed_to_based_permission(["score", "indication", "redeem"]).has_permission(
            request, None
        )
        self.assertFalse(permission_rslt)

    @mock.patch(
        "jsm_user_services.drf_tools.permissions.IsUserAllowedToPermission._validate_request_against_user_service"
    )
    def test_should_return_true_if_user_data_has_redeem_as_not_allowed_to_parameter_but_there_is_no_restriction(
        self, mocked_validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": ["redeem"]}

        # full match default equals to False
        permission_rslt = self.allowed_to_based_permission(["indication", "score"]).has_permission(request, None)
        self.assertTrue(permission_rslt)

        permission_rslt = self.allowed_to_based_permission(["redee", "score", "indication"]).has_permission(
            request, None
        )
        self.assertTrue(permission_rslt)

    @mock.patch(
        "jsm_user_services.drf_tools.permissions.IsUserAllowedToPermission._validate_request_against_user_service"
    )
    def test_should_return_true_if_user_data_is_empty_or_do_not_return_not_allowed_to_parameter(
        self, mocked_validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked_validate_request_against_user_service.return_value = {}
        permission_rslt = self.allowed_to_based_permission(["redeem"]).has_permission(request, None)
        self.assertTrue(permission_rslt)

        mocked_validate_request_against_user_service.return_value = {"not_allowed_to": []}
        permission_rslt = self.allowed_to_based_permission(["redeem"]).has_permission(request, None)
        self.assertTrue(permission_rslt)

        mocked_validate_request_against_user_service.return_value = {"another_parameter": ["redeem"]}
        permission_rslt = self.allowed_to_based_permission(["redeem"]).has_permission(request, None)
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.perform_recaptcha_validation")
    def test_should_perform_a_request_to_google_if_captcha_key_is_on_request(self, mocked_google_service):

        request = mock.MagicMock()
        request.headers = {}
        # since the key is on the body, the request to google should be made
        request.data = {"g_recaptcha_response": "some cool value"}

        self.google_permission_recaptcha.has_permission(request, None)

        assert mocked_google_service.called

    @mock.patch("jsm_user_services.drf_tools.permissions.perform_recaptcha_validation")
    def test_should_not_perform_a_request_to_google_if_captcha_key_is_not_on_request(self, mocked_google_service):

        request = mock.MagicMock()

        # since there's no g_recaptcha_response key in both header and body, it shouldn't perform the request to google
        request.headers = {}
        request.data = {}

        self.google_permission_recaptcha.has_permission(request, None)

        assert not mocked_google_service.called

    @mock.patch("jsm_user_services.drf_tools.permissions.perform_recaptcha_validation")
    def test_should_raise_if_key_is_not_on_request_but_exception_is_defined(self, mocked_google_service):
        class TestException(APIException):
            pass

        # noinspection PyTypeChecker
        google_permission_with_another_exception = GoogleRecaptchaPermission(TestException())

        request = mock.MagicMock()

        # since there's no g_recaptcha_response key in both header and body, it shouldn't perform the request to google
        request.headers = {}
        request.data = {}

        with pytest.raises(TestException):
            google_permission_with_another_exception.has_permission(request, None)

        assert not mocked_google_service.called

    @mock.patch("jsm_user_services.drf_tools.permissions.perform_recaptcha_validation")
    def test_should_override_header_key_if_the_same_key_is_present_on_body_too(self, mocked_google_service):

        recaptcha_value_on_body = "some value that will be used"
        request = mock.MagicMock()
        # since the priority is the body data and the key is present on both header and body, the header
        # should be ignored
        request.headers = {"g-recaptcha-response": "some value that will be ignored :("}
        request.data = {"g-recaptcha-response": recaptcha_value_on_body}

        self.google_permission_recaptcha.has_permission(request, None)

        mocked_google_service.assert_called_once_with(recaptcha_value_on_body)

    @mock.patch("jsm_user_services.drf_tools.permissions.perform_recaptcha_validation")
    def test_should_get_the_underscore_key_instead_of_dash_if_both_keys_are_present(self, mocked_google_service):

        recaptcha_value_on_key_with_underscore = "some value that will be used"
        request = mock.MagicMock()
        request.headers = {}
        # since the priority is the underscore (g_recaptcha_response) key and both underscore (g_recaptcha_response) and
        # dash(g-recaptcha-response) are present, the dash key should be ignored
        request.data = {
            "g-recaptcha-response": "some value that will be ignored :(",
            "g_recaptcha_response": recaptcha_value_on_key_with_underscore,
        }

        self.google_permission_recaptcha.has_permission(request, None)

        mocked_google_service.assert_called_once_with(recaptcha_value_on_key_with_underscore)

    def test_permission_class_should_succeed_if_google_verification_succeeded(self):

        request = mock.MagicMock()
        request.headers = {}
        request.data = {
            "g_recaptcha_response": "some cool value",
        }

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.post(
                "https://www.google.com/recaptcha/api/siteverify",
                status_code=200,
                json={"success": True, "score": 0.8},
            )

            assert self.google_permission_recaptcha.has_permission(request, None)
            assert mocked_request.called

    def test_permission_class_should_fail_if_google_verification_failed(self):

        request = mock.MagicMock()
        request.headers = {}
        request.data = {
            "g_recaptcha_response": "some cool value",
        }

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.post(
                "https://www.google.com/recaptcha/api/siteverify",
                status_code=200,
                json={"success": False},
            )

            assert not self.google_permission_recaptcha.has_permission(request, None)
            assert mocked_request.called

    def test_permission_class_should_raise_if_google_verification_failed_and_api_exception_is_initialized(self):
        class TestException(APIException):
            pass

        google_permission_with_api_exception = GoogleRecaptchaPermission(TestException())

        request = mock.MagicMock()
        request.headers = {}
        request.data = {
            "g_recaptcha_response": "some cool value",
        }

        with Mocker(real_http=True) as mocker:
            with pytest.raises(TestException):
                mocked_request = mocker.post(
                    "https://www.google.com/recaptcha/api/siteverify",
                    status_code=200,
                    json={"success": False},
                )

                google_permission_with_api_exception.has_permission(request, None)
            assert mocked_request.called

    def test_permission_class_should_raise_if_google_verification_failed_and_api_exception_is_not_initialized(self):
        class TestException(APIException):
            pass

        google_permission_with_api_exception = GoogleRecaptchaPermission(TestException)

        request = mock.MagicMock()
        request.headers = {}
        request.data = {
            "g_recaptcha_response": "some cool value",
        }

        with Mocker(real_http=True) as mocker:
            with pytest.raises(TestException):
                mocked_request = mocker.post(
                    "https://www.google.com/recaptcha/api/siteverify",
                    status_code=200,
                    json={"success": False},
                )

                google_permission_with_api_exception.has_permission(request, None)
            assert mocked_request.called

    def test_permission_class_should_not_raise_if_google_verification_failed_and_non_api_exception_is_initialized(self):
        class TestException(Exception):
            pass

        # noinspection PyTypeChecker
        google_permission_with_another_exception = GoogleRecaptchaPermission(TestException())

        request = mock.MagicMock()
        request.headers = {}
        request.data = {
            "g_recaptcha_response": "some cool value",
        }

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.post(
                "https://www.google.com/recaptcha/api/siteverify",
                status_code=200,
                json={"success": False},
            )

            assert not google_permission_with_another_exception.has_permission(request, None)
            assert mocked_request.called

    def test_permission_class_should_not_raise_if_google_verification_failed_and_non_api_exception_is_not_initialized(
        self,
    ):
        class TestException(Exception):
            pass

        # noinspection PyTypeChecker
        google_permission_with_another_exception = GoogleRecaptchaPermission(TestException)

        request = mock.MagicMock()
        request.headers = {}
        request.data = {
            "g_recaptcha_response": "some cool value",
        }

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.post(
                "https://www.google.com/recaptcha/api/siteverify",
                status_code=200,
                json={"success": False},
            )

            assert not google_permission_with_another_exception.has_permission(request, None)
            assert mocked_request.called

    def test_should_allow_use_google_permission_initialized(self):
        permission = GoogleRecaptchaPermission()

        class FakeViewSet(ViewSet):

            permission_classes = [permission]

        fake_view_set = FakeViewSet()
        assert fake_view_set.get_permissions() == [permission]

    def test_should_allow_use_allowed_to_permission_initialized(self):
        permission = IsUserAllowedToPermission(["redeem", "score"], full_match=False)

        class FakeViewSet(ViewSet):

            permission_classes = [permission]

        fake_view_set = FakeViewSet()
        assert fake_view_set.get_permissions() == [permission]

    def test_permission_class_should_fail_if_google_verification_score_is_less_than_default(self):

        request = mock.MagicMock()
        request.headers = {}
        request.data = {
            "g_recaptcha_response": "some cool value",
        }

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.post(
                "https://www.google.com/recaptcha/api/siteverify",
                status_code=200,
                json={"success": True, "score": 0.7},
            )

            assert self.google_permission_recaptcha.has_permission(request, None) is False
            assert mocked_request.called



from typing import Any

from jsm_user_services import local_threading


def add_to_local_threading(key: str, value: Any):
    setattr(local_threading, key, value)


def remove_from_local_threading(key: str):
    try:
        delattr(local_threading, key)
    except:
        pass


def get_from_local_threading(key: str, default_value=None) -> Any:
    return getattr(local_threading, key, default_value)



import logging
import os
from importlib import import_module
from typing import Optional

from flask import Flask
from flask import request

from jsm_user_services.support.auth_jwt import get_bearer_authorization_token
from jsm_user_services.support.local_threading_utils import add_to_local_threading
from jsm_user_services.support.local_threading_utils import remove_from_local_threading
from jsm_user_services.support.string_utils import get_first_value_from_comma_separated_string

logger = logging.getLogger(__name__)

ip_address_meta_key = os.getenv("GUNICORN_IP_ADDRESS_HEADER", "x-original-forwarded-for")


class JsmJwtService:
    def __init__(self, app: Optional[Flask] = None):
        """
        Initialize the JsmJwtService middleware if the flask application (`app`) is given. Otherwise, you'll
        have the option to initialize it later.

        Examples:
        - Initialize as soon as possible:

        ```
        from flask import Flask
        from jsm_user_services.flask.middleware import JsmJwtService

        app = Flask(__name__)
        JsmJwtService(app)
        ```

        - Postpone initialization:

        ```
        from flask import Flask
        from jsm_user_services.flask.middleware import JsmJwtService

        app, middleware = Flask(__name__), JsmJwtService()
        middleware.init_app(app)
        ```
        """

        self.app = app

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        @app.before_request
        def add_infos_to_local_threading():
            add_to_local_threading("authorization_token", self._get_jwt_token_from_request())
            add_to_local_threading("user_ip", self._get_user_ip_from_request())
            add_to_local_threading("user_session_id", self._get_user_session_id_from_request())

        @app.after_request
        def remove_infos_from_local_threading(response: Flask.response_class) -> Flask.response_class:
            remove_from_local_threading("authorization_token")
            remove_from_local_threading("user_ip")
            remove_from_local_threading("user_session_id")
            return response

    @staticmethod
    def _get_jwt_token_from_request() -> Optional[str]:
        """
        Extracts JWT token from a Flask request object.
        """
        authorization_value = request.headers.get("Authorization", "")
        return get_bearer_authorization_token(authorization_value)

    @staticmethod
    def _get_user_ip_from_request() -> Optional[str]:
        """
        Retrieve the user ip that made this request from Flask request object

        When running a service behind Akamai or other CDN solutions, it is expected that this header might contain a string with multiple IPs (comma separated values). In this case, the user's public IP that originated the request is considered to be the first one of this list.
        For reference: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-For
        """
        return get_first_value_from_comma_separated_string(request.headers.get(ip_address_meta_key))

    @staticmethod
    def _get_user_session_id_from_request() -> Optional[str]:
        bearer_token = JsmJwtService._get_jwt_token_from_request()
        if not bearer_token:
            return None

        user = import_module("jsm_user_services.services.user")
        return user.get_session_id_from_bearer_token(bearer_token)



from unittest.mock import Mock

from django.test import TestCase
from jsm_user_services.middleware import JsmJwtService
from jsm_user_services.services.user import get_session_id_from_bearer_token
from jsm_user_services.support.local_threading_utils import get_from_local_threading
from jsm_user_services.support.string_utils import get_first_value_from_comma_separated_string


def _build_mocked_request(user_ip: str) -> Mock:
    request = Mock()
    request.META = {
        "REQUEST_METHOD": "POST",
        "HTTP_X_ORIGINAL_FORWARDED_FOR": user_ip,
        "HTTP_AUTHORIZATION": "Bearer abc",
    }
    request.path = "/fake/url/"
    request.session = {}
    return request


class TestMiddleware(TestCase):
    def setUp(self):
        self.user_ip = "8.8.8.8"
        self.user_ip_comma_separated = "7.7.7.7, 8.8.8.8, 9.9.9.9"

        self.middleware = JsmJwtService()

        self.request = _build_mocked_request(self.user_ip)
        self.request_comma_separated_user_ip = _build_mocked_request(self.user_ip_comma_separated)

    def test_should_get_user_ip_from_request(self):
        response = self.middleware.process_request(self.request)
        self.assertIsNone(response)

        self.assertEqual(get_from_local_threading("user_ip"), self.user_ip)

    def test_should_get_user_ip_from_request_when_an_ip_list_is_given(self):
        expected_user_ip = get_first_value_from_comma_separated_string(self.user_ip_comma_separated)
        self.assertEqual(expected_user_ip, "7.7.7.7")

        response = self.middleware.process_request(self.request_comma_separated_user_ip)
        self.assertIsNone(response)
        self.assertEqual(get_from_local_threading("user_ip"), expected_user_ip)

    def test_should_get_user_session_id_from_request(self):
        response, expected_session_id = (
            self.middleware.process_request(self.request),
            get_session_id_from_bearer_token("abc"),
        )

        self.assertIsNone(response)
        self.assertEqual(get_from_local_threading("user_session_id"), expected_session_id)

    def test_should_retrieve_user_session_id_as_none_when_request_is_made_without_bearer_token(self):
        del self.request.META["HTTP_AUTHORIZATION"]
        self.middleware.process_request(self.request)

        self.assertIsNone(get_from_local_threading("user_session_id"))

        self.request.META["HTTP_AUTHORIZATION"] = "Token abc"
        self.middleware.process_request(self.request)

        self.assertIsNone(get_from_local_threading("user_session_id"))

    def test_should_replace_user_session_id_after_request_made_with_bearer_token(self):
        self.middleware.process_request(self.request)

        self.assertIsNotNone(get_from_local_threading("user_session_id"))

        del self.request.META["HTTP_AUTHORIZATION"]
        self.middleware.process_request(self.request)

        self.assertIsNone(get_from_local_threading("user_session_id"))

    def test_should_replace_authorization_token_after_request_made_with_bearer_token(self):
        self.middleware.process_request(self.request)

        self.assertIsNotNone(get_from_local_threading("authorization_token"))

        del self.request.META["HTTP_AUTHORIZATION"]
        self.middleware.process_request(self.request)

        self.assertIsNone(get_from_local_threading("authorization_token"))

    def test_should_replace_user_ip_after_request_made_with_expected_ip_header(self):
        self.middleware.process_request(self.request)

        self.assertIsNotNone(get_from_local_threading("authorization_token"))

        del self.request.META["HTTP_X_ORIGINAL_FORWARDED_FOR"]
        self.middleware.process_request(self.request)

        self.assertIsNone(get_from_local_threading("user_ip"))



from unittest import TestCase

from requests import exceptions

from jsm_user_services.support.http_utils import convert_header_to_meta_key
from jsm_user_services.support.http_utils import request


class TestContracts(TestCase):
    def test_should_raise_timeout(self):
        with request() as r:
            # check https://stackoverflow.com/a/100859 for the reason of "http://10.255.255.1/"
            self.assertRaises(exceptions.Timeout, r.get, "http://10.255.255.1")

    def test_should_convert_header_to_meta_key(self):
        header1 = "x-original-forwarded-for"
        header2 = "authentication"
        header3 = "x-forwarded-for"

        self.assertEqual(convert_header_to_meta_key(header1), "HTTP_X_ORIGINAL_FORWARDED_FOR")
        self.assertEqual(convert_header_to_meta_key(header2), "HTTP_AUTHENTICATION")
        self.assertEqual(convert_header_to_meta_key(header3), "HTTP_X_FORWARDED_FOR")



from typing import Optional

from jsm_user_services import settings
from jsm_user_services.exception import MissingRequiredConfiguration


def get_from_settings_or_raise_missing_config(name: str, default: Optional[str] = None) -> str:
    """
    Tries to retrieve the value from attribute in settings. If it was not possible to retrieve the value, raises.
    """

    # getattr can return None when the attribute exists and the value is None
    value = getattr(settings, name, default) or default
    if value is None:
        raise MissingRequiredConfiguration(f"The variable {name} is missing")

    return value



from unittest.mock import patch

from django.test import TestCase
from requests_mock import Mocker

from jsm_user_services.exception import MissingRequiredConfiguration
from jsm_user_services.services.google import perform_recaptcha_validation


class GoogleServiceTest(TestCase):
    @patch("jsm_user_services.support.settings_utils.getattr", side_effect=["test", None])
    def test_should_raise_if_missing_secret(self, mocked_getattr):

        with self.assertRaises(MissingRequiredConfiguration):
            perform_recaptcha_validation("")

    @patch("jsm_user_services.support.settings_utils.getattr", side_effect=["http://test.com", "test", "0.8"])
    def test_should_perform_request_if_both_vars_exists(self, mocked_getattr):

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.post("http://test.com", status_code=200, json={"success": True, "score": 0.8})
            assert perform_recaptcha_validation("")
            assert mocked_request.called



import pytest

from jsm_user_services.decorators.lgpd_utils import AnonymizeLogic
from jsm_user_services.exception import DecoratorWrongUsage


class TestAnonymizeName:
    def test_should_raise_if_name_is_not_a_string(self):
        with pytest.raises(DecoratorWrongUsage):
            AnonymizeLogic.anonymize_name(123)

    def test_should_return_none_if_name_is_none(self):
        assert AnonymizeLogic.anonymize_name(None) is None

    def test_should_return_first_name_and_last_name_if_name_is_complete(self):
        assert AnonymizeLogic.anonymize_name("Lucas Gigek Carvalho") == "Lucas C"

    def test_should_return_only_first_name_if_there_is_no_last_name(self):
        assert AnonymizeLogic.anonymize_name("Lucas") == "Lucas"


class TestAnonymizeCpf:
    def test_should_raise_if_cpf_is_not_a_string(self):
        with pytest.raises(DecoratorWrongUsage):
            AnonymizeLogic.anonymize_cpf(12345678910)

    def test_should_none_if_cpf_is_none(self):
        assert AnonymizeLogic.anonymize_cpf(None) is None

    def test_should_return_anonymized_cpf(self):
        assert AnonymizeLogic.anonymize_cpf("12345678910") == "*******8910"


class TestAnonymizePhone:
    def test_should_raise_if_phone_is_not_a_string(self):
        with pytest.raises(DecoratorWrongUsage):
            AnonymizeLogic.anonymize_phone(5511912345678)

    def test_should_return_none_if_phone_is_none(self):
        assert AnonymizeLogic.anonymize_phone(None) is None

    def test_should_return_anonymized_phone(self):
        assert AnonymizeLogic.anonymize_phone("5511912345678") == "*********5678"


class TestAnonymizeEmail:
    def test_should_raise_if_email_is_not_a_string(self):
        with pytest.raises(DecoratorWrongUsage):
            AnonymizeLogic.anonymize_email(123)

    def test_should_return_none_if_email_is_none(self):
        assert AnonymizeLogic.anonymize_email(None) is None

    def test_should_return_anonymized_phone(self):
        assert (
            AnonymizeLogic.anonymize_email("lucas.carvalho@juntossomosmais.com.br")
            == "l*************@juntossomosmais.com.br"
        )

    def test_should_return_only_username_if_there_is_no_domain(self):
        assert AnonymizeLogic.anonymize_email("lucas.carvalho") == "l*************"


class TestAnonymizeAddressShouldNotBeReturned:
    def test_should_return_none(self):
        assert AnonymizeLogic.anonymize_address_should_not_be_returned("Av. Gomes de Carvalho") is None


class TestAnonymizeAddressShouldBeReturned:
    def test_should_return_the_same_data(self):
        assert AnonymizeLogic.anonymize_address_should_be_returned("São Paulo") == "São Paulo"



from jsm_user_services import local_threading
from jsm_user_services.support.local_threading_utils import add_to_local_threading
from jsm_user_services.support.local_threading_utils import get_from_local_threading
from jsm_user_services.support.local_threading_utils import remove_from_local_threading


def test_should_set_local_threading_attribute_to_the_given_value():
    add_to_local_threading("new_attribute", "abc")

    assert hasattr(local_threading, "new_attribute")
    assert local_threading.new_attribute == "abc"

    del local_threading.new_attribute


def test_should_get_from_local_threading_attribute_value():
    local_threading.new_attribute = "abc"

    assert get_from_local_threading("new_attribute") == "abc"

    del local_threading.new_attribute


def test_should_remove_from_local_threading_attribute():
    local_threading.new_attribute = "abc"

    remove_from_local_threading("new_attribute")

    assert not hasattr(local_threading, "new_attribute")



import logging
from typing import Callable
from typing import Dict
from typing import List

from requests import Response

from jsm_user_services.exception import EmailValidationError
from jsm_user_services.support.email_utils import perform_email_validators
from jsm_user_services.support.email_utils import validate_format_email
from jsm_user_services.support.http_utils import get_response_body
from jsm_user_services.support.http_utils import request
from jsm_user_services.support.settings_utils import get_from_settings_or_raise_missing_config

logger = logging.getLogger(__name__)


DEFAULT_VALIDATORS_FUNCTION = [validate_format_email]


def email_is_valid(
    email: str,
    use_callback_function_validator: bool = False,
    functions_to_validate_email: List[Callable] = DEFAULT_VALIDATORS_FUNCTION,
):
    """
    Returns if the given e-mail is valid
        Parameters:
            `email (str)`: the e-mail that will be validated
            `use_callback_function_validator (bool)`: indicates if some validators must be considered if the `status_code`
            is different from 200 (success), otherwise raises an exception
            `functions_to_validate_email (List[Callable])`: a list of validators that will be considered if the `status_code`
            is different from 200 (success) and the `use_callback_function_validator` is `True`

    """
    EVEREST_API_KEY = get_from_settings_or_raise_missing_config("EVEREST_API_KEY")
    EVEREST_API_URL = get_from_settings_or_raise_missing_config(
        "EVEREST_API_HOST", "https://api.everest.validity.com/api/2.0/validation/addresses"
    )

    with request(status_force_list=(429, 500, 502, 503, 504)) as r:
        url: str = f"{EVEREST_API_URL}/{email}"
        headers: Dict = {"X-API-KEY": EVEREST_API_KEY}

        response: Response = r.get(url, headers=headers)
        payload: Dict = get_response_body(response)
        if response.status_code == 400:
            status: str = check_response_payload_error(payload)
            logger.warning(f"{response.status_code} - {status}")
            return False
        if response.status_code != 200:
            status: str = check_response_payload_error(payload)
            logger.warning(f"{response.status_code} - {status}")
            if use_callback_function_validator:
                return perform_email_validators(
                    functions_to_validate_email=functions_to_validate_email,
                    email=email
                )
            raise EmailValidationError
        logger.info(f"Everest API Response: {payload}")
        return check_response_payload(payload)


def check_response_payload(payload: Dict) -> bool:
    """
    Checks if the response received from Everest indicated if the email is valid
        Example of success payload:
        {
            "meta": {},
            "results": {
                "category": "valid",
                "status": "valid",
                "name": "Valid",
                "definition": "The email has a valid account associated with it.",
                "reasons": [],
                "risk": "low",
                "recommendation": "send",
                "address": "ayron41@gmail.com",
                "diagnostics": {
                "role_address": false,
                "disposable": false,
                "typo": false
                }
            }
        }
    """
    results: Dict = payload.get("results") or {}
    status: str = results.get("status")
    if status:
        return status != "invalid"
    return False


def check_response_payload_error(payload: Dict) -> str:
    """
    Check the error received from Everest
        Example of error payload:
        ```{"status": "Bad Request: Invalid email address."}```
    """
    if payload:
        status: str = payload.get("status")
        return status
    return "Everest is out!"



import pytest

from jsm_user_services.decorators.lgpd import anonymize_sensible_data
from jsm_user_services.decorators.lgpd_utils import AnonymizeLogic
from jsm_user_services.exception import DecoratorWrongUsage


class TestAnonymizeSensibleData:
    def test_should_raise_if_not_initialized_and_no_param_is_defined(self):
        @anonymize_sensible_data
        def my_method():
            pass

        with pytest.raises(
            TypeError, match=r"wrapper\(\) missing 1 required positional argument: 'method_that_uses_the_decorator'"
        ):
            my_method()

    def test_should_raise_if_initialized_but_no_param_is_defined(self):
        with pytest.raises(
            TypeError, match=r"anonymize_sensible_data\(\) missing 1 required positional argument: 'to_be_anonymized'"
        ):
            # it fails before the method's execution
            @anonymize_sensible_data()
            def my_method():
                pass

    def test_should_raise_if_param_is_not_a_dict(self):
        with pytest.raises(DecoratorWrongUsage, match="Please, pass a dict as param."):
            # it fails before the method's execution
            @anonymize_sensible_data("test")
            def my_method():
                pass

    def test_should_raise_if_key_is_not_a_string(self):
        with pytest.raises(
            DecoratorWrongUsage, match=r"('The keys must be a str, got key %s of type %s', 1, <class 'int'>)"
        ):
            # it fails before the method's execution
            @anonymize_sensible_data({"correct": AnonymizeLogic.NAME, 1: AnonymizeLogic.CPF})
            def my_method():
                pass

    def test_should_raise_if_value_is_not_from_enum(self):
        with pytest.raises(
            DecoratorWrongUsage, match=r"('The values must be a AnonymizeLogic, got value %s', 'some random logic')"
        ):
            # it fails before the method's execution
            @anonymize_sensible_data({"correct": AnonymizeLogic.NAME, "incorrect": "some random logic"})
            def my_method():
                pass

    def test_should_raise_if_method_does_not_return_a_dict(self):
        @anonymize_sensible_data({"correct": AnonymizeLogic.NAME})
        def my_method():
            return "me ajuda cara"

        with pytest.raises(
            DecoratorWrongUsage,
            match=r"('The method must return a dict or a list, but returned type %s', <class 'str'>)",
        ):
            my_method()

    def test_should_raise_if_key_does_not_exist(self):
        @anonymize_sensible_data({"some_key": AnonymizeLogic.NAME})
        def my_method():
            return {}

        with pytest.raises(DecoratorWrongUsage, match=r"('Key not found %s', 'some_key')"):
            my_method()

    def test_should_not_raise_if_key_does_not_exist_but_param_is_false(self):
        @anonymize_sensible_data({"some_key": AnonymizeLogic.NAME}, False)
        def my_method():
            return {}

        assert my_method() == {}

    def test_should_user_pattern_from_param(self):
        @anonymize_sensible_data({"test1xptotest2": AnonymizeLogic.NAME}, traverse_dict_pattern="xpto")
        def my_method():
            return {"test1": {"test2": "Lucas Gigek Carvalho"}}

        assert my_method() == {"test1": {"test2": "Lucas C"}}

    def test_should_anonymize_data_according_to_param(self):
        @anonymize_sensible_data(
            {
                "name": AnonymizeLogic.NAME,
                "cpf": AnonymizeLogic.CPF,
                "phone": AnonymizeLogic.PHONE,
                "email": AnonymizeLogic.EMAIL,
                "city": AnonymizeLogic.ADDRESS_CITY,
                "state": AnonymizeLogic.ADDRESS_STATE,
                "neighborhood": AnonymizeLogic.ADDRESS_NEIGHBORHOOD,
                "street": AnonymizeLogic.ADDRESS_STREET,
                "number": AnonymizeLogic.ADDRESS_NUMBER,
                "complement": AnonymizeLogic.ADDRESS_COMPLEMENT,
                "postal_code": AnonymizeLogic.ADDRESS_POSTAL_CODE,
            }
        )
        def my_method():
            return {
                "name": "Lucas Gigek Carvalho",
                "cpf": "12345678910",
                "phone": "5511912345678",
                "email": "lucas.carvalho@juntossomosmais.com.br",
                "city": "São Paulo",
                "state": "SP",
                "neighborhood": "Vila Olímpia",
                "street": "Av. Gomes de Carvalho",
                "number": "1666",
                "complement": "8o andar",
                "postal_code": "04547-006",
            }

        assert my_method() == {
            "name": "Lucas C",
            "cpf": "*******8910",
            "phone": "*********5678",
            "email": "l*************@juntossomosmais.com.br",
            "city": "São Paulo",
            "state": "SP",
            "neighborhood": None,
            "street": None,
            "number": None,
            "complement": None,
            "postal_code": None,
        }

    def test_should_anonymize_data_on_nested_dicts(self):
        @anonymize_sensible_data(
            {
                "address__city": AnonymizeLogic.ADDRESS_CITY,
                "address__state": AnonymizeLogic.ADDRESS_STATE,
                "address__neighborhood": AnonymizeLogic.ADDRESS_NEIGHBORHOOD,
                "address__street": AnonymizeLogic.ADDRESS_STREET,
                "address__number": AnonymizeLogic.ADDRESS_NUMBER,
                "address__complement": AnonymizeLogic.ADDRESS_COMPLEMENT,
                "address__postal_code": AnonymizeLogic.ADDRESS_POSTAL_CODE,
            }
        )
        def my_method():
            return {
                "address": {
                    "city": "São Paulo",
                    "state": "SP",
                    "neighborhood": "Vila Olímpia",
                    "street": "Av. Gomes de Carvalho",
                    "number": "1666",
                    "complement": "8o andar",
                    "postal_code": "04547-006",
                }
            }

        assert my_method() == {
            "address": {
                "city": "São Paulo",
                "state": "SP",
                "neighborhood": None,
                "street": None,
                "number": None,
                "complement": None,
                "postal_code": None,
            }
        }

    def test_should_anonymize_data_on_lists(self):
        @anonymize_sensible_data(
            {
                "addresses__data__city": AnonymizeLogic.ADDRESS_CITY,
                "addresses__data__state": AnonymizeLogic.ADDRESS_STATE,
                "addresses__data__neighborhood": AnonymizeLogic.ADDRESS_NEIGHBORHOOD,
                "addresses__data__street": AnonymizeLogic.ADDRESS_STREET,
                "addresses__data__number": AnonymizeLogic.ADDRESS_NUMBER,
                "addresses__data__complement": AnonymizeLogic.ADDRESS_COMPLEMENT,
                "addresses__data__postal_code": AnonymizeLogic.ADDRESS_POSTAL_CODE,
            }
        )
        def my_method():
            return {
                "addresses": [
                    {
                        "type": "work",
                        "data": {
                            "city": "São Paulo",
                            "state": "SP",
                            "neighborhood": "Vila Olímpia",
                            "street": "Av. Gomes de Carvalho",
                            "number": "1666",
                            "complement": "8o andar",
                            "postal_code": "04547-006",
                        },
                    },
                    {
                        "type": "personal",
                        "data": {
                            "city": "São Paulo",
                            "state": "SP",
                            "neighborhood": "Um bairro honesto",
                            "street": "Uma rua mais honesta ainda",
                            "number": "2",
                            "complement": None,
                            "postal_code": "03874-000",
                        },
                    },
                ]
            }

        assert my_method() == {
            "addresses": [
                {
                    "type": "work",
                    "data": {
                        "city": "São Paulo",
                        "state": "SP",
                        "neighborhood": None,
                        "street": None,
                        "number": None,
                        "complement": None,
                        "postal_code": None,
                    },
                },
                {
                    "type": "personal",
                    "data": {
                        "city": "São Paulo",
                        "state": "SP",
                        "neighborhood": None,
                        "street": None,
                        "number": None,
                        "complement": None,
                        "postal_code": None,
                    },
                },
            ]
        }

    def test_should_anonymize_user_data_from_ishtar(self):
        @anonymize_sensible_data(
            {
                "cpf": AnonymizeLogic.CPF,
                "name": AnonymizeLogic.NAME,
                "emails__email": AnonymizeLogic.EMAIL,
                "phones__number": AnonymizeLogic.PHONE,
                "username": AnonymizeLogic.CPF,
                "addresses__city": AnonymizeLogic.ADDRESS_CITY,
                "addresses__state": AnonymizeLogic.ADDRESS_STATE,
                "addresses__number": AnonymizeLogic.ADDRESS_NUMBER,
                "addresses__street": AnonymizeLogic.ADDRESS_STREET,
                "addresses__district": AnonymizeLogic.ADDRESS_NEIGHBORHOOD,
                "addresses__complement": AnonymizeLogic.ADDRESS_COMPLEMENT,
                "addresses__postal_code": AnonymizeLogic.ADDRESS_POSTAL_CODE,
            }
        )
        def my_method():
            return {
                "id": "b74ad26a-e80a-4756-bdc9-1207fd3c62e9",
                "cpf": "44053216044",
                "name": "Proprietário Automação sHAZRscr",
                "roles": ["owner"],
                "emails": [{"type": "personal", "email": "diego.ferreira@juntossomosmais.com.br"}],
                "gender": "male",
                "phones": [
                    {"type": "mobile", "number": "5511981656209"},
                    {"type": "landline", "number": "551155380499"},
                ],
                "status": "active",
                "mediums": ["whatsapp"],
                "birthday": "1990-01-01",
                "username": "44053216044",
                "addresses": [
                    {
                        "city": "São Paulo",
                        "type": "main",
                        "state": "SP",
                        "number": "474",
                        "street": "Rua João Fugulin",
                        "country": "BRA",
                        "district": "Jardim Germânia",
                        "complement": "3564572",
                        "postal_code": "05849340",
                    },
                    {
                        "city": "São Paulo",
                        "type": "shipping",
                        "state": "SP",
                        "number": "474",
                        "street": "Rua João Fugulin",
                        "country": "BRA",
                        "district": "Jardim Germânia",
                        "complement": "3564572",
                        "postal_code": "05849340",
                    },
                ],
            }

        assert my_method() == {
            "id": "b74ad26a-e80a-4756-bdc9-1207fd3c62e9",
            "cpf": "*******6044",
            "name": "Proprietário s",
            "roles": ["owner"],
            "emails": [{"type": "personal", "email": "d*************@juntossomosmais.com.br"}],
            "gender": "male",
            "phones": [
                {"type": "mobile", "number": "*********6209"},
                {"type": "landline", "number": "********0499"},
            ],
            "status": "active",
            "mediums": ["whatsapp"],
            "birthday": "1990-01-01",
            "username": "*******6044",
            "addresses": [
                {
                    "city": "São Paulo",
                    "type": "main",
                    "state": "SP",
                    "number": None,
                    "street": None,
                    "country": "BRA",
                    "district": None,
                    "complement": None,
                    "postal_code": None,
                },
                {
                    "city": "São Paulo",
                    "type": "shipping",
                    "state": "SP",
                    "number": None,
                    "street": None,
                    "country": "BRA",
                    "district": None,
                    "complement": None,
                    "postal_code": None,
                },
            ],
        }

    def test_should_not_raise_if_list_contains_different_data_structures(self):
        @anonymize_sensible_data({"users__name": AnonymizeLogic.NAME})
        def my_method():
            return {"users": ["Baraka", "Johnny Cage", {"name": "Liu Kang"}]}

        assert my_method() == {"users": ["Baraka", "Johnny Cage", {"name": "Liu K"}]}

    def test_should_anonymize_if_dict_has_nested_lists(self):
        @anonymize_sensible_data({"users__name": AnonymizeLogic.NAME})
        def my_method():
            return {
                "users": [[{"name": "Baraka"}, {"name": "Liu Kang"}], [{"name": "Kung Lao"}, {"name": "Johnny Cage"}]]
            }

        assert my_method() == {
            "users": [[{"name": "Baraka"}, {"name": "Liu K"}], [{"name": "Kung L"}, {"name": "Johnny C"}]]
        }

    def test_should_anonymize_if_param_is_a_list(self):
        @anonymize_sensible_data({"name": AnonymizeLogic.NAME})
        def my_method():
            return [{"name": "Baraka"}, {"name": "Liu Kang"}, {"name": "Kung Lao"}, {"name": "Johnny Cage"}]

        assert my_method() == [{"name": "Baraka"}, {"name": "Liu K"}, {"name": "Kung L"}, {"name": "Johnny C"}]

    def test_should_work_if_method_receives_a_param(self):
        @anonymize_sensible_data({"name": AnonymizeLogic.NAME})
        def my_method(should_return, *args, **kwargs):
            return should_return

        assert my_method({"name": "Liu Kang"}) == {"name": "Liu K"}
        assert my_method(should_return={"name": "Liu Kang"}) == {"name": "Liu K"}
        assert my_method({"name": "Liu Kang"}, useless_param_that_shouldnt_impact="test") == {"name": "Liu K"}



import pytest
from django.conf import LazySettings as DjangoSettings
from flask import Flask
from flask.config import Config as FlaskSettings
from pytest_mock import MockFixture

from jsm_user_services import settings
from jsm_user_services.exception import ApplicationSettingsNotFound
from jsm_user_services.settings import get_setting_from_app


@pytest.fixture
def flask_app():
    app = Flask(__name__)

    app.config["USER_API_HOST"] = "http://flask-app"

    with app.app_context():
        yield


@pytest.fixture
def no_django(mocker: MockFixture):
    mocker.patch("jsm_user_services.settings.get_django_settings", return_value=None)


@pytest.fixture
def no_flask(mocker: MockFixture):
    mocker.patch("jsm_user_services.settings.get_flask_settings", return_value=None)


def test_should_correctly_retrieve_django_settings():
    django_settings = settings.get_django_settings()

    assert isinstance(django_settings, DjangoSettings)
    assert django_settings.USER_API_HOST == "http://ishtar-gate.dev.juntossomosmaisi.com.br/api/v1"


def test_should_correctly_retrieve_flask_settings(flask_app: None):
    flask_settings = settings.get_flask_settings()

    assert isinstance(flask_settings, FlaskSettings)
    assert flask_settings.get("USER_API_HOST") == "http://flask-app"


def test_should_correctly_retrieve_setting_from_flask_app(flask_app: None, no_django: None):
    assert get_setting_from_app("USER_API_HOST") == "http://flask-app"


def test_should_correctly_retrieve_settings_from_django_app(no_flask: None):
    assert get_setting_from_app("USER_API_HOST") == "http://ishtar-gate.dev.juntossomosmaisi.com.br/api/v1"


def test_should_raise_exception_when_application_settings_isnt_found(no_django: None):
    with pytest.raises(ApplicationSettingsNotFound):
        get_setting_from_app("USER_API_HOST")



import logging
import os
from typing import Optional

from django.http import HttpRequest

from jsm_user_services.services.user import get_session_id_from_bearer_token
from jsm_user_services.support.auth_jwt import get_bearer_authorization_token
from jsm_user_services.support.http_utils import convert_header_to_meta_key
from jsm_user_services.support.local_threading_utils import add_to_local_threading
from jsm_user_services.support.local_threading_utils import remove_from_local_threading
from jsm_user_services.support.string_utils import get_first_value_from_comma_separated_string

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object

logger = logging.getLogger(__name__)

ip_address_meta_key = convert_header_to_meta_key(os.getenv("GUNICORN_IP_ADDRESS_HEADER", "x-original-forwarded-for"))


class JsmJwtService(MiddlewareMixin):
    def process_request(self, request):
        add_to_local_threading("authorization_token", self._get_jwt_token_from_request(request))
        add_to_local_threading("user_ip", self._get_user_ip_from_request(request))
        add_to_local_threading("user_session_id", self._get_user_session_id_from_request(request))

    def process_response(self, request, response):
        remove_from_local_threading("authorization_token")
        remove_from_local_threading("user_ip")
        return response

    @staticmethod
    def _get_jwt_token_from_request(request: HttpRequest) -> Optional[str]:
        """
        Extracts JWT token from a Django request object.
        """
        authorization_value = request.META.get("HTTP_AUTHORIZATION", "")
        return get_bearer_authorization_token(authorization_value)

    @staticmethod
    def _get_user_ip_from_request(request: HttpRequest) -> Optional[str]:
        """
        Retrieve the user ip that made this request from Django HttpRequest object

        When running a service behind Akamai or other CDN solutions, it is expected that this header might contain a string with multiple IPs (comma separated values). In this case, the user's public IP that originated the request is considered to be the first one of this list.
        For reference: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-For
        """
        return get_first_value_from_comma_separated_string(request.META.get(ip_address_meta_key, None))

    @staticmethod
    def _get_user_session_id_from_request(request: HttpRequest) -> Optional[str]:
        bearer_token = JsmJwtService._get_jwt_token_from_request(request)
        if not bearer_token:
            return None
        return get_session_id_from_bearer_token(bearer_token)



from enum import Enum
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from jsm_user_services.exception import DecoratorWrongUsage


class AnonymizeLogic(Enum):
    """
    An Enum that allows decorator's user decide which logic should be used.
    """

    @staticmethod
    def anonymize_name(name: Optional[str]):
        """
        Anonymize strategy for "name" type. It uses the following logic:
        >>> anonymize_name("Lucas Gigek Carvalho")
            Lucas C

        In order to do it, the name must be a string.
        """

        if name is None:
            return None

        if not isinstance(name, str):
            raise DecoratorWrongUsage("Name must be a string, but got %s", type(name))

        splitted_name = name.split(" ")
        if len(splitted_name) == 1:
            return name

        return f"{splitted_name.pop(0)} {splitted_name.pop()[slice(1)]}"

    @staticmethod
    def anonymize_cpf(cpf: Optional[str]):
        """
        Anonymize strategy for "CPF" type. It uses the following logic:
        >>> anonymize_cpf("12345678910")
            *******8910

        In order to do it, the cpf must be a string.
        """

        if cpf is None:
            return None

        if not isinstance(cpf, str):
            raise DecoratorWrongUsage("CPF must be a string, but got %s", type(cpf))

        return cpf[slice(-4, None)].rjust(len(cpf), "*")

    @staticmethod
    def anonymize_phone(phone: Optional[str]):
        """
        Anonymize strategy for "phone" type. It uses the following logic:
        >>> anonymize_phone("5511912345678")
            *********5678

        In order to do it, the phone must be a string.
        """

        if phone is None:
            return None

        if not isinstance(phone, str):
            raise DecoratorWrongUsage("Phone must be a string, but got %s", type(phone))

        return phone[slice(-4, None)].rjust(len(phone), "*")

    @staticmethod
    def anonymize_email(email: Optional[str]):
        """
        Anonymize strategy for "email" type. It uses the following logic:
        >>> anonymize_email("lucas.carvalho@juntossomosmais.com.br")
            l*************@juntossomosmais.com.br

        In order to do it, the email must be a string.
        """

        if email is None:
            return None

        if not isinstance(email, str):
            raise DecoratorWrongUsage("Email must be a string, but got %s", type(email))

        splitted_email = email.split("@")

        username = splitted_email[0][slice(1)].ljust(len(splitted_email[0]), "*")
        if len(splitted_email) == 1:
            return username

        return f"{username}@{splitted_email[1]}"

    @staticmethod
    def anonymize_address_should_not_be_returned(address_data: Optional[str]):
        """
        Anonymize strategy for "address" type which shouldn't be returned. It uses the following logic:
        >>> anonymize_address_should_not_be_returned("Av. Gomes de Carvalho")
            None
        """

        return None

    @staticmethod
    def anonymize_address_should_be_returned(address_data: Optional[str]):
        """
        Anonymize strategy for "address" type which should be returned. It uses the following logic:
        >>> anonymize_address_should_be_returned("São Paulo")
            "São Paulo"
        """

        return address_data

    NAME = anonymize_name
    CPF = anonymize_cpf
    PHONE = anonymize_phone
    EMAIL = anonymize_email
    ADDRESS_CITY = anonymize_address_should_be_returned
    ADDRESS_STATE = anonymize_address_should_be_returned
    ADDRESS_NEIGHBORHOOD = anonymize_address_should_not_be_returned
    ADDRESS_STREET = anonymize_address_should_not_be_returned
    ADDRESS_NUMBER = anonymize_address_should_not_be_returned
    ADDRESS_COMPLEMENT = anonymize_address_should_not_be_returned
    ADDRESS_POSTAL_CODE = anonymize_address_should_not_be_returned


def traverse_within_dict_overriding_data(
    path: str,
    original_data: Union[Dict, List],
    anonymize_logic: Callable,
    raise_if_key_not_found: bool,
    traverse_dict_pattern: str,
):
    """
    Method that navigates through the dict and modifies the necessary keys.
    """

    path_to_traverse = path.split(traverse_dict_pattern)

    # by doing this, all changes made on "current_key" will go to "original_data" as well
    current_key = original_data

    while len(path_to_traverse) > 0:

        if isinstance(current_key, list):
            # iterating on each item to deal with them recursively
            for item in current_key:
                traverse_within_dict_overriding_data(
                    traverse_dict_pattern.join(path_to_traverse),
                    item,
                    anonymize_logic,
                    raise_if_key_not_found,
                    traverse_dict_pattern,
                )
            # data was already changed on inner iterations, there's no need to continue the loop
            break

        next_key = path_to_traverse.pop(0)

        # If it's not a dict, it's not possible to use ".get" on it.
        if not isinstance(current_key, dict):
            continue

        # If it's the last key, the only thing that must be done is change it's data.
        # Also, it's necessary to check that the key exists, in order to not creat an additional key.
        if len(path_to_traverse) == 0 and current_key.get(next_key):
            # calling method that will anonymize data
            current_key[next_key] = anonymize_logic(current_key.get(next_key))
            # data was already changed, there's no need to continue the loop
            break

        try:
            current_key = current_key[next_key]
        except KeyError:
            if raise_if_key_not_found:
                raise DecoratorWrongUsage("Key not found %s", next_key)

    return original_data



DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "testdatabase"}}

DEBUG = True

INSTALLED_APPS = ["jsm_user_services"]

SECRET_KEY = "hewf!63(!t1j-9=8v%_tcg1_p1zz!5aywb%^b$vow17se!8y84"

USER_API_HOST = "http://ishtar-gate.dev.juntossomosmaisi.com.br/api/v1"
USER_API_TOKEN = "9f0350491d4355e3d4a966712d110d3e7d62a9c7"

JSM_USER_SERVICE_HTTP_TIMEOUT = 1e-6  # 0.000001

# DRF permissions
JSM_USER_SERVICES_DRF_APPEND_USER_DATA = True
JSM_USER_SERVICES_DRF_REQUEST_USER_DATA_ATTR_NAME = "user_data"
GOOGLE_RECAPTCHA_SECRET_KEY = "some_cool_secret"



import pytest
from flask.testing import FlaskClient

from jsm_user_services.services.user import get_session_id_from_bearer_token


def test_should_retrieve_authorization_token_that_is_set_on_local_threading(flask_test_client: FlaskClient):
    response = flask_test_client.get("/authorization-token", headers={"Authorization": "Bearer abc"})

    assert response.status_code == 200
    assert response.json == {"auth_token": "abc"}


@pytest.mark.parametrize("ips_string,expected_ip", [("8.8.8.8", "8.8.8.8"), ("7.7.7.7, 8.8.8.8, 9.9.9.9", "7.7.7.7")])
def test_should_retrieve_user_ip_that_is_set_on_local_threading(
    flask_test_client: FlaskClient, ips_string: str, expected_ip: str
):
    response = flask_test_client.get("/my-ip", headers={"x-original-forwarded-for": ips_string})

    assert response.status_code == 200
    assert response.json == {"ip": expected_ip}


def test_should_retrieve_user_session_id_that_is_set_on_local_threading(flask_test_client: FlaskClient):
    response = flask_test_client.get("/session-id", headers={"Authorization": "Bearer abc"})

    assert response.status_code == 200
    assert response.json == {"session_id": get_session_id_from_bearer_token("abc")}



import pytest
from flask_log_request_id.ctx_fetcher import MultiContextRequestIdFetcher
from pytest_mock import MockFixture
from request_id_django_log import local_threading

from jsm_user_services.exception import RequestIDModuleNotFound
from jsm_user_services.support.request_id import get_current_request_id_callable
from jsm_user_services.support.request_id import import_module_otherwise_none


@pytest.fixture
def no_flask_request_id_module(mocker: MockFixture):
    def mock_only_if_flask_log_request_id_module(module_to_import: str):
        if module_to_import == "flask_log_request_id":
            return None
        return import_module_otherwise_none(module_to_import)

    mocker.patch(
        "jsm_user_services.support.request_id.import_module_otherwise_none",
        side_effect=mock_only_if_flask_log_request_id_module,
    )


@pytest.fixture
def no_django_request_id_module(mocker: MockFixture):
    def mock_only_if_django_request_id_module(module_to_import: str):
        if module_to_import == "request_id_django_log.request_id":
            return None
        return import_module_otherwise_none(module_to_import)

    mocker.patch(
        "jsm_user_services.support.request_id.import_module_otherwise_none",
        side_effect=mock_only_if_django_request_id_module,
    )


@pytest.fixture
def no_request_id_module_found(mocker: MockFixture):
    mocker.patch("jsm_user_services.support.request_id.import_module_otherwise_none", return_value=None)


def test_should_successfully_retrieve_current_request_id_for_django_app(no_flask_request_id_module: None):
    local_threading.request_id, current_request_id = ("django", get_current_request_id_callable())

    django_request_id = current_request_id()

    assert django_request_id == local_threading.request_id

    del local_threading.request_id


def test_should_succesfully_retrieve_current_request_id_for_flask_app(no_django_request_id_module: None):
    current_request_id = get_current_request_id_callable()

    assert isinstance(current_request_id, MultiContextRequestIdFetcher)


def test_should_raise_exception_when_request_id_module_is_not_found(no_request_id_module_found: None):
    with pytest.raises(RequestIDModuleNotFound):
        get_current_request_id_callable()



from typing import Callable

from jsm_user_services.exception import RequestIDModuleNotFound
from jsm_user_services.support.import_utils import import_module_otherwise_none


def get_current_request_id_callable() -> Callable:
    django_request_id_module, flask_request_id_module = (
        import_module_otherwise_none("request_id_django_log.request_id"),
        import_module_otherwise_none("flask_log_request_id"),
    )

    if django_request_id_module is not None:
        return django_request_id_module.current_request_id

    if flask_request_id_module is not None:
        return flask_request_id_module.current_request_id

    raise RequestIDModuleNotFound


current_request_id = get_current_request_id_callable()



"""
Module with useful user permissions classes meant to be used with Django Rest Framework.
"""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import requests
from django.conf import settings
from jsm_user_services.drf_tools.helpers import AllowClassBehaviorAsFunction
from jsm_user_services.drf_tools.helpers import is_exception_related_to_api_exception
from jsm_user_services.exception import IncorrectTypePermissionConfiguration
from jsm_user_services.services.google import perform_recaptcha_validation
from jsm_user_services.services.user import get_user_data_from_server
from rest_framework import permissions
from rest_framework.exceptions import APIException
from rest_framework.request import Request

logger = logging.getLogger(__name__)


class JSMUserBasePermission(permissions.BasePermission):
    """
    Base class for JSM user permissions. Implements methods for validating requests against
    the user micro service so that any type of permission can be carried out (role, status, etc).
    """

    APPEND_USER_DATA = getattr(settings, "JSM_USER_SERVICES_DRF_APPEND_USER_DATA", True)
    USER_DATA_ATTR_NAME = getattr(settings, "JSM_USER_SERVICES_DRF_REQUEST_USER_DATA_ATTR_NAME", "user_data")
    USER_SERVICE_NOT_AUTHORIZED_CODES = (401, 403, 404)

    @classmethod
    def _retrieve_user_data(cls, request: Request) -> Dict:
        try:
            user_data = getattr(request, cls.USER_DATA_ATTR_NAME)
        except AttributeError:
            user_data = get_user_data_from_server()

        return user_data

    @classmethod
    def _validate_request_against_user_service(cls, request: Request, append_user_data_to_request: bool = True) -> Dict:
        """
        Gets valid user_data from the User micro service.
        """
        try:
            user_data: Dict = cls._retrieve_user_data(request)

            if append_user_data_to_request:
                setattr(request, cls.USER_DATA_ATTR_NAME, user_data)

        except requests.exceptions.HTTPError as e:
            if e.response.status_code in cls.USER_SERVICE_NOT_AUTHORIZED_CODES:
                return {}

            raise

        return user_data

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Abstract method that must be implemented by children classes.
        """
        raise NotImplementedError("This method must be overridden!")


class StatusBasedPermission(JSMUserBasePermission):
    """
    Base class for the status-based permissions. Implements methods for validating requests against
    the user micro service and an utility method that asserts that the user has the appropriate status.
    """

    @classmethod
    def _validate_user_status(cls, request: Request, allowed_status: List[str]):
        """
        Validates if the user has the appropriate status against the allowed status.
        """
        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )

        if "status" not in user_data:
            return False

        return user_data["status"] in allowed_status


class RoleBasedPermission(JSMUserBasePermission):
    """
    Base class for the role-based permissions. Implements methods for validating requests against
    the user micro service and an utility method that asserts that the user has the appropriate role.
    """

    @classmethod
    def _validate_user_role(cls, request: Request, allowed_roles: List[str]):
        """
        Validates if the user has the appropriate role against the allowed roles.
        """
        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )
        user_roles: List[str] = user_data.get("roles", [])

        if not user_roles:
            return False

        return any([user_role in allowed_roles for user_role in user_roles])


class ProfessionalUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'professional'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is a professional.
        """
        return self._validate_user_role(request, allowed_roles=["professional"])


class EmployeeOrManagerUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'employee' or 'manager'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is either an employee or manager.
        """
        return self._validate_user_role(request, allowed_roles=["employee", "manager"])


class EmployeeOnlyUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'employee'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an employee.
        """
        return self._validate_user_role(request, allowed_roles=["employee"])


class OwnerUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'owner'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an owner.
        """
        return self._validate_user_role(request, allowed_roles=["owner"])


class IndustrySellerUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'industry-seller'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an industry seller.
        """
        return self._validate_user_role(request, allowed_roles=["industry-seller"])


class AnyLoggedUserPermission(RoleBasedPermission):
    """
    Permission that allows users to access the requested resource regardless
    of their roles as long as they have a VALID, NON EXPIRED JSM JWT.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is just logged in with a valid jwt (non expired).
        """
        return self._validate_user_role(request, allowed_roles=["professional", "employee", "manager", "owner"])


class ActiveUserPermission(StatusBasedPermission):
    """
    Permission that allows only users which the status is 'active'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user status is 'active'.
        """
        return self._validate_user_status(request, allowed_status=["active"])


class PendingValidationUserPermission(StatusBasedPermission):
    """
    Permission that allows only users which the status is 'pending-validation'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user status is 'pending-validation'.
        """
        return self._validate_user_status(request, allowed_status=["pending-validation"])


class GoogleRecaptchaPermission(permissions.BasePermission, AllowClassBehaviorAsFunction):
    """
    A permission class which checks if the request is authorized by Google Recaptcha V3.

    In order to use it, the request must have the key "g_recaptcha_response" or "g-recaptcha-response" on the request's
    body or header. Otherwise the request won't be authorized without even checking it on Google.

    Just be aware that a header with underscore is not allowed. So do not use the key "g_recaptcha_response"
    on the header.
    """

    def __init__(self, exception_in_case_of_failed: Optional[Union[APIException, Type[APIException]]] = None):
        """
        Sets the default value for "exception_in_case_of_failed_verification" property. It will be used to return
        a custom response, in case of this permission fails.
        This exception must inherit APIException, from rest_framework.exceptions, otherwise it will be ignored.
        """

        self.exception_in_case_of_failed_verification = exception_in_case_of_failed

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Retrieves the Google data on the request and then performs a request to Google, in order to check if the
        received data is valid or not.
        """

        g_recaptcha_response = self._retrieve_g_recaptcha_response(request)
        use_received_exception = is_exception_related_to_api_exception(self.exception_in_case_of_failed_verification)

        # preventing an unnecessary request to Google, since key is not defined
        if not g_recaptcha_response:
            if use_received_exception:
                raise self.exception_in_case_of_failed_verification
            return False

        google_response: bool = perform_recaptcha_validation(g_recaptcha_response)

        if not google_response and use_received_exception:
            # raising an APIException that will return the customized response
            raise self.exception_in_case_of_failed_verification

        return google_response

    @classmethod
    def _retrieve_g_recaptcha_response(cls, request: Request) -> Optional[str]:
        """
        Lookup on the request's body and header in order to retrieve the recaptcha key.

         If the same key is present on header and body, the header data will be overridden by the body value.
         Also, if both keys are present, the priority is "g_recaptcha_response" over "g-recaptcha-response".
        """

        expected_keys = ["g_recaptcha_response", "g-recaptcha-response"]

        for key in expected_keys:
            key_value = request.data.get(key) or request.headers.get(key)
            if key_value:
                return key_value


class IsUserAllowedToPermission(JSMUserBasePermission, AllowClassBehaviorAsFunction):
    """
    A permission class which checks if user is authorized to execute an action based on 'not_allowed_to' parameter.
    Implements methods for validating requests against the user micro service and an utility method that asserts
    the user has the appropriate permission for to do, or not to do, a determined action.

    Example:
        class SomeAwesomeView():
            permission_classes = [
                IsUserAllowedToPermission(
                    ["redeem", "score"], full_match=False, exception_in_case_of_failed=CustomExceptionNotAllowedTo
                )
            ]

        In this example, the custom exception should be implemented into your own project.
    """

    def __init__(
        self,
        not_allowed_to: List[str],
        full_match: bool = False,
        exception_in_case_of_failed: Optional[Union[APIException, Type[APIException]]] = None,
    ):
        """
        Sets the default value for "exception_in_case_of_failed_verification" property. It will be used to return
        a custom response, in case of this permission fails. This exception must inherit APIException,
        from rest_framework.exceptions, otherwise it will be ignored.

        Full match property indicates that user data must contains all the 'not_allowed_to' defined parameters.
        On the other hand, if full match is equal to False, the permission class checks
        if user data contains any of the 'not_allowed_to' parameters.
        """

        self.exception_in_case_of_failed_verification = exception_in_case_of_failed
        self.full_match = full_match

        if not isinstance(not_allowed_to, list):
            logger.error(
                f"Permission class 'IsUserAllowedToPermission' expected 'List[str]', got  {type(not_allowed_to)}"
            )
            raise IncorrectTypePermissionConfiguration(f"Expected type 'List[str]', got {type(not_allowed_to)} instead")

        self.not_allowed_to = not_allowed_to

    @classmethod
    def _verify_if_user_should_be_allowed(
        cls, request: Request, not_allowed_actions: List[str], full_match: bool
    ) -> bool:
        """
        Validates 'not_allowed_to' user parameter against the 'not_allowed_actions' permission class parameter.
        """

        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )
        user_not_allowed_to: List[str] = user_data.get("not_allowed_to", [])

        set_from_not_allowed_actions = set(not_allowed_actions)
        set_from_user_not_allowed_to = set(user_not_allowed_to)
        intersected_actions = set_from_not_allowed_actions.intersection(set_from_user_not_allowed_to)

        if not (user_not_allowed_to or intersected_actions):
            return True

        if full_match:
            return len(intersected_actions) != len(set_from_not_allowed_actions)
        return not bool(intersected_actions)

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user hasn't an action restriction against the 'not_allowed_to' permission class parameter.
        """

        use_received_exception: bool = is_exception_related_to_api_exception(
            self.exception_in_case_of_failed_verification
        )

        is_allowed: bool = self._verify_if_user_should_be_allowed(
            request, not_allowed_actions=self.not_allowed_to, full_match=self.full_match,
        )

        if use_received_exception and not is_allowed:
            raise self.exception_in_case_of_failed_verification

        return is_allowed



import os

from setuptools import find_packages
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="jsm_user_services",
    version="1.1.0",
    description="Middleware to intercept JWT auth token and more utils functions",
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    author="Juntos Somos Mais",
    author_email="labs@juntossomosmais.com.br",
    license="MIT",
    url="https://github.com/juntossomosmais/jsm-user-service",
    packages=find_packages(),
    install_requires=["PyJWT==2.1.0", "requests"],
    extras_require={
        "drf": ["djangorestframework", "request-id-django-log"],
        "flask": ["flask", "flask-log-request-id"],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Framework :: Django :: 2.1",
        "Environment :: Web Environment",
        "Natural Language :: Portuguese (Brazilian)",
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import hashlib
from importlib import import_module
from typing import List
from typing import Optional

from jsm_user_services.support.auth_jwt import get_decoded_jwt_token
from jsm_user_services.support.http_utils import get_response_body
from jsm_user_services.support.http_utils import request
from jsm_user_services.support.local_threading_utils import get_from_local_threading
from jsm_user_services.support.request_id import current_request_id


def jwt_has_required_roles(jwt_required_roles: List[str], assert_all=True) -> bool:
    """
    Asserts that the jwt token has all the required roles.
    """
    jsm_token_data = get_jsm_user_data_from_jwt()

    if not jsm_token_data:
        return False

    try:
        jwt_token_roles = jsm_token_data["roles"]
    except KeyError:
        return False

    if not jwt_token_roles:
        return False

    if assert_all:
        return all([(jwt_token_role in jwt_required_roles) for jwt_token_role in jwt_token_roles])

    return any([(jwt_token_role in jwt_required_roles) for jwt_token_role in jwt_token_roles])


def current_jwt_token() -> Optional[str]:
    return get_from_local_threading("authorization_token")


def get_jsm_token() -> Optional[str]:
    token = current_jwt_token()
    if token:
        return get_decoded_jwt_token(token)["jsm_identity"]

    return None


def get_ltm_token() -> Optional[str]:
    token = current_jwt_token()
    if token:
        return get_decoded_jwt_token(token)["yuntiandu"]

    return None


def get_jsm_user_data_from_jwt() -> Optional[dict]:
    token = get_jsm_token()

    if token:
        return get_decoded_jwt_token(token)

    return None


def get_user_email_from_jwt() -> Optional[str]:
    user_data = get_jsm_user_data_from_jwt()
    if user_data:
        return user_data.get("email")

    return None


def get_user_id_from_jwt() -> Optional[str]:
    user_data = get_jsm_user_data_from_jwt()
    if user_data:
        return user_data.get("uid")

    return None


def get_user_access_as_id_from_jwt() -> Optional[str]:
    user_data = get_jsm_user_data_from_jwt()
    if user_data:
        return user_data.get("fid", None)

    return None


def get_user_data_from_server() -> dict:

    current_token = current_jwt_token()
    headers = {"REQUEST-ID": current_request_id()}
    settings = import_module("jsm_user_services.settings")
    user_url = settings.USER_API_HOST

    if current_token:
        headers["Authorization"] = f"Bearer {current_jwt_token()}"

    with request() as r:
        response = r.get(f"{user_url}/users/me/", headers=headers)
        return get_response_body(response)


def get_user_data_from_cpf(cpf: str) -> dict:

    current_token = current_jwt_token()
    headers = {"REQUEST-ID": current_request_id()}
    settings = import_module("jsm_user_services.settings")
    user_url = settings.USER_API_HOST

    if current_token:
        headers["Authorization"] = f"Bearer {current_jwt_token()}"

    with request() as r:
        response = r.get(f"{user_url}/users/search?cpf={cpf}", headers=headers)
        return get_response_body(response)


def get_user_data_from_id(user_id: str) -> dict:

    current_token = current_jwt_token()
    headers = {"REQUEST-ID": current_request_id()}
    settings = import_module("jsm_user_services.settings")
    user_url = settings.USER_API_HOST

    if current_token:
        headers["Authorization"] = f"Bearer {current_jwt_token()}"

    with request() as r:
        response = r.get(f"{user_url}/users/{user_id}/", headers=headers)

        return get_response_body(response)


def get_cpf_from_jwt() -> Optional[str]:

    email = get_user_email_from_jwt()

    return email.split("@")[0] if email else None


def is_retail_user(user_id: str) -> bool:
    settings = import_module("jsm_user_services.settings")
    user_api_token = settings.USER_API_TOKEN
    user_url = settings.USER_API_HOST

    headers = {"REQUEST-ID": current_request_id(), "Authorization": f"Token {user_api_token}"}

    with request() as r:
        response = r.get(f"{user_url}/users/search/?user_id_ref={user_id}&is_retail_user=True", headers=headers)
        response_content = get_response_body(response)
    return response_content["count"] == 1


def get_user_ip() -> Optional[str]:
    return get_from_local_threading("user_ip")


def get_session_id_from_bearer_token(bearer_token: str) -> str:
    return hashlib.sha1(bearer_token.encode("utf-8")).hexdigest()


def get_user_session_id() -> Optional[str]:
    return get_from_local_threading("user_session_id")



import pytest
from flask import Flask
from flask import jsonify
from flask.testing import FlaskClient

from jsm_user_services.flask import middleware
from jsm_user_services.support.local_threading_utils import get_from_local_threading


@pytest.fixture(autouse=True)
def flask_app() -> Flask:
    app = Flask(__name__)

    app.config.setdefault("USER_API_HOST", "http://flask-app")
    middleware.JsmJwtService(app)

    @app.route("/authorization-token", methods=["GET"])
    def retrieve_authorization_token():
        return jsonify({"auth_token": get_from_local_threading("authorization_token")})

    @app.route("/my-ip", methods=["GET"])
    def retrieve_user_ip():
        return jsonify({"ip": get_from_local_threading("user_ip")})

    @app.route("/session-id", methods=["GET"])
    def retrieve_user_session_id():
        return jsonify({"session_id": get_from_local_threading("user_session_id")})

    return app


@pytest.fixture
def flask_test_client(flask_app: Flask) -> FlaskClient:
    flask_app.testing = True
    return flask_app.test_client()



from contextlib import suppress
from importlib import import_module
from typing import Optional

from jsm_user_services.exception import ApplicationSettingsNotFound
from jsm_user_services.support.import_utils import import_module_otherwise_none


def get_django_settings():
    django_conf = import_module_otherwise_none("django.conf")
    if django_conf and django_conf.settings.configured:
        return django_conf.settings
    return None


def get_flask_settings():
    with suppress(ImportError, RuntimeError):
        flask = import_module("flask")
        return flask.current_app.config


def get_setting_from_app(setting_name: str, default_value: Optional[str] = None):
    django_settings, flask_settings = get_django_settings(), get_flask_settings()
    if django_settings is not None:
        return getattr(django_settings, setting_name, default_value)

    if flask_settings is not None:
        return flask_settings.get(setting_name, default_value)

    raise ApplicationSettingsNotFound


USER_API_HOST = get_setting_from_app("USER_API_HOST")
USER_API_TOKEN = get_setting_from_app("USER_API_TOKEN")
JSM_USER_SERVICE_HTTP_TIMEOUT = get_setting_from_app("JSM_USER_SERVICE_HTTP_TIMEOUT", "30")
GOOGLE_RECAPTCHA_URL = get_setting_from_app("GOOGLE_RECAPTCHA_URL")
GOOGLE_RECAPTCHA_SECRET_KEY = get_setting_from_app("GOOGLE_RECAPTCHA_SECRET_KEY")
GOOGLE_RECAPTCHA_SCORE_THRESHOLD = get_setting_from_app("GOOGLE_RECAPTCHA_SCORE_THRESHOLD")
EVEREST_API_KEY = get_setting_from_app("EVEREST_API_KEY")
EVEREST_API_HOST = get_setting_from_app("EVEREST_API_HOST")



from types import ModuleType

from jsm_user_services.support.import_utils import import_module_otherwise_none


def test_should_return_none_when_import_error_is_raised():
    assert import_module_otherwise_none("unknown_module") is None


def test_should_return_imported_module_when_it_is_possible():
    imported_module = import_module_otherwise_none("jsm_user_services")

    assert isinstance(imported_module, ModuleType)



import re
from typing import Optional

import jwt


def get_decoded_jwt_token(jwt_token: str) -> dict:
    """
    Gets a decoded JWT token.
    """

    return jwt.decode(jwt_token, algorithms=["HS256"], options={"verify_signature": False})


def get_bearer_authorization_token(authorization_value: str) -> Optional[str]:
    """
    Retrieve a bearer authorization token from an Authorization header value.

    It expects the header value to be something on the lines of: "Bearer token".

    Examples:
    - get_bearer_authorization_token("Bearer token") # returns "token"
    - get_bearer_authorization_token("bearer token") # returns None
    - get_bearer_authorization_token("Token token") # returns None
    - get_bearer_authorization_token("whatever") # returns None
    """
    match = re.match("Bearer", authorization_value)

    if not match:
        return None

    auth_type_beginning = match.span()[1]
    jwt_token = authorization_value[auth_type_beginning:].strip()

    return jwt_token



import logging

from jsm_user_services.support.http_utils import request
from jsm_user_services.support.settings_utils import get_from_settings_or_raise_missing_config

logger = logging.getLogger(__name__)


def perform_recaptcha_validation(g_recaptcha_response: str) -> bool:
    """
    Performs a request to Google in order to validate the reCAPTCHA.
    For more details, check: https://developers.google.com/recaptcha/docs/verify
    """

    google_recaptcha_url = get_from_settings_or_raise_missing_config(
        "GOOGLE_RECAPTCHA_URL", "https://www.google.com/recaptcha/api/siteverify"
    )
    google_recaptcha_secret_key = get_from_settings_or_raise_missing_config("GOOGLE_RECAPTCHA_SECRET_KEY")

    min_score_threshold = float(get_from_settings_or_raise_missing_config("GOOGLE_RECAPTCHA_SCORE_THRESHOLD", "0.8"))

    logger.debug("Performing request to Google to check if recaptcha is valid")


    data = {"response": g_recaptcha_response, "secret": google_recaptcha_secret_key}

    with request() as r:
        resp = r.post(google_recaptcha_url, data=data)
    
    result_json = resp.json()

    logger.info(f"Status code: {resp.status_code}; min_score_threshold: {min_score_threshold}, body: {result_json}")

    if resp.status_code != 200:
        logger.info(f"GoogleRecaptcha response {g_recaptcha_response} status: {resp.status_code} failed: {result_json}")
        return False

    if result_json.get("success") is not True:
        return False

    if result_json.get("score", 0) < min_score_threshold:
        return False

    return True
